# written by Danna Du
# adapted from TECO_matrix Fortran code written by Chris Lu
# need packages numpy and XlsxWriter installed to run code

import numpy as np
import xlsxwriter

# Create a new Excel file
workbook = xlsxwriter.Workbook('output1.xlsx', {'nan_inf_to_errors': True})    # 'nan_inf_to_errors' allows
#                    'output1' is the name of output file                         conversion of NAN/INF in Python to
#                                                                                 #NUM! and #DIV/0! error in Excel,
# add a worksheet                                                                 default is False
worksheet = workbook.add_worksheet()
# format cells: center-align
cell_format = workbook.add_format()
cell_format.set_align('center')

# set initial day of year (doy) and year
doy = 1
year = 1

# from teco.nml
nyr_sim = 10000            # number of years in simulation
nyr_forc = 10              # number of years in forcing
startyr = 1                # starting year
dt = 86400                 # time step (seconds), daily 1 day = 86400 seconds

use_matrix = True          # True: Turn on matrix representation, False: Conventional representation
is_SASU_spinup = True      # True: Turn on SASU spin up, False: Turn off SASU spin up
is_diagnostic = True       # True: Calculate diagnostic variables output, False: Not calculate diagnostic variables
is_write_yearly = True     # True: Write output yearly, False: Write output daily

frac_foliage_to_metlit = 0.7123         # C transfer coefficient (foliage to metabolic litter)
frac_foliage_to_strlit = 0.2877         # C transfer coefficient (foliage to structural litter
frac_woody_to_strlit = 1                # C transfer coefficient (wood to structural litter)
frac_metlit_to_micsoil = 0.45           # C transfer coefficient (metabolic litter to soil microbial)
frac_strlit_to_micsoil = 0.275          # C transfer coefficient (structural litter to soil microbial)
frac_slowsoil_to_micsoil = 0.42         # C transfer coefficient (slow soil to soil microbial)
frac_passsoil_to_micsoil = 0.45         # C transfer coefficient (passive soil to soil microbial)
frac_micsoil_to_slowsoil = 0.296        # C transfer coefficient (soil microbial to slow soil)
frac_strlit_to_slowsoil = 0.275         # C transfer coefficient (structural litter to slow soil)
frac_micsoil_to_passsoil = 0.004        # C transfer coefficient (soil microbial to passive soil)
frac_slowsoil_to_passsoil = 0.03        # C transfer coefficient (slow soil to passive soil)

# the following numbers indicate the position of matrices (zero-indexed):
ifoliage = 0
iwoody = 1
imetlit = 2
istrlit = 3
imicsoil = 4
islowsoil = 5
ipasssoil = 6

# number of pools (determines size of matrices)
npools = 7

# 1-dimensional matrix (1:npools=7)
dC_dt = [0 for d in range(npools)]          # creates the following matrix: [0, 0, 0, 0, 0, 0, 0]
#                                            (matrices represented as lists in python or as arrays by numpy)
# ↑ in this case, use place-holding zeros to create initial matrix structure -- values change accordingly as code runs

k = [0.00176, 0.000100104, 0.021468, 0.000845, 0.008534, 8.976e-005, 0.00000154782]   # ← turnover rate per day of
#                                                                                         pools: foliage, wood,
#                                                                                         metabolic litter, structural
#                                                                                         litter, soil microbial, slow
#                                                                                         soil, passive soil
Calloc = [0.45, 0.55, 0, 0, 0, 0, 0]    # ← NPP allocation fraction to foliage, wood, metabolic litter, structural
#                                           litter, soil microbial, slow soil, passive soil


for i in range(len(k)):
    k[i] = k[i] / 3600 / 24


def C_input():
    """resets the value of Cinput to 0.00002245 at the start of each simulation year (value from Cinput_Spinup)"""
    global Cinput
    Cinput = 0.00002245


# 1-dimensional matrix (1:npools)
Cpool = [0 for p in range(npools)]
matrix_B = [0 for b in range(npools)]

# 2-dimensional matrices (1:npools, 1:npools)
matrix_A = [[0.0 for column in range(npools)]for row in range(npools)]
matrix_K = [[0.0 for col in range(npools)]for r in range(npools)]
# creates the following matrix (list of lists):
# [[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
#  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
#  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
#  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
#  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
#  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
#  [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]]

# making diagonal values -1 in matrix_A
for i in range(len(matrix_A)):
    matrix_A[i][i] = -1


def open_io_file():
    """adjusts column widths and writes headers in first row of output Excel worksheet"""
    #    ↓ start column:end column, column width
    worksheet.set_column('A:B', 7)          # making columns narrower
    worksheet.set_column('C:N', 16)         # making columns wider
    col = 0             # start writing in first column
    headers = ['Year', 'Doy', 'Cinput(gC/s)', 'Foliage(gC/m2)', 'Woody(gC/m2)', 'Met_litter(gC/m2)',
               'Str_litter(gC/m2)', 'Fast_soil(gC/m2)', 'Slow_soil(gC/m2)', 'Pass_soil(gC/m2)', 'I(gC/m2/s)',
               'Residence time(yr)', 'Capacity(gC/m2)', 'Potential(gC/m2)']           # list of header names
    for header in headers:
        #           ↓ row, column, text, center-align
        worksheet.write(0, col, header, cell_format)
        col += 1        # go to next column


def tick_time():
    """changes doy and year accordingly and prints out the year"""
    global year, doy
    doy += 1
    if doy > 365:
        print('year', 2*'\t', year)
        year += 1
        doy -= 365


def is_endof_year():
    """determines whether or not it is the end of a year"""
    if doy == 365:
        is_end_of_year = True
    else:
        is_end_of_year = False
    return is_end_of_year


def is_startof_year():
    """determines whether or not it is the start of a year"""
    if doy == 1:
        is_start_of_year = True
    else:
        is_start_of_year = False
    return is_start_of_year


def is_endof_forcing():
    """determines whether or not it is the end of a forcing cycle"""
    is_end_of_forcing = (year - startyr + 1) % nyr_forc
    if is_end_of_forcing == 0 and is_endof_year():
        is_end_of_forcing = True
    else:
        is_end_of_forcing = False
    return is_end_of_forcing


def inverse(a):
    """
    takes the inverse of the input matrix
    :param a: input matrix
    :return: inverse of a
    """
    c = np.linalg.inv(a)
    return c


def update_ecosysC():
    global dC_dt
    if use_matrix:
        matrix_B[ifoliage] = Calloc[ifoliage]
        matrix_B[iwoody] = Calloc[iwoody]

        matrix_K[ifoliage][ifoliage] = k[ifoliage]
        matrix_K[iwoody][iwoody] = k[iwoody]
        matrix_K[imetlit][imetlit] = k[imetlit]
        matrix_K[istrlit][istrlit] = k[istrlit]
        matrix_K[imicsoil][imicsoil] = k[imicsoil]
        matrix_K[islowsoil][islowsoil] = k[islowsoil]
        matrix_K[ipasssoil][ipasssoil] = k[ipasssoil]

        matrix_A[imetlit][ifoliage] = frac_foliage_to_metlit
        matrix_A[istrlit][ifoliage] = frac_foliage_to_strlit
        matrix_A[istrlit][iwoody] = frac_woody_to_strlit
        matrix_A[imicsoil][imetlit] = frac_metlit_to_micsoil
        matrix_A[imicsoil][istrlit] = frac_strlit_to_micsoil
        matrix_A[imicsoil][islowsoil] = frac_slowsoil_to_micsoil
        matrix_A[imicsoil][ipasssoil] = frac_passsoil_to_micsoil
        matrix_A[islowsoil][imicsoil] = frac_micsoil_to_slowsoil
        matrix_A[islowsoil][istrlit] = frac_strlit_to_slowsoil
        matrix_A[ipasssoil][imicsoil] = frac_micsoil_to_passsoil
        matrix_A[ipasssoil][islowsoil] = frac_slowsoil_to_passsoil

        dC_dt = np.add((np.multiply(Cinput, matrix_B)), (np.matmul((np.matmul(matrix_A, matrix_K)), Cpool)))

        for i in range(len(Cpool)):
            Cpool[i] += (dC_dt[i] * dt)

    elif not use_matrix:
        # plant
        dC_dt[ifoliage] = Cinput * Calloc[ifoliage] - k[ifoliage] * Cpool[ifoliage]
        dC_dt[iwoody] = Cinput * Calloc[iwoody] - k[iwoody] * Cpool[iwoody]

        # litter
        dC_dt[imetlit] = Cpool[ifoliage] * k[ifoliage] * frac_foliage_to_metlit - Cpool[imetlit] * k[imetlit]
        dC_dt[istrlit] = Cpool[ifoliage] * k[ifoliage] * frac_foliage_to_strlit \
                         + Cpool[iwoody] * k[iwoody] * frac_woody_to_strlit \
                         - Cpool[istrlit] * k[istrlit]          # the backslashes \ indicate a line break (no meaning)

        # soil
        dC_dt[imicsoil] = Cpool[imetlit] * k[imetlit] * frac_metlit_to_micsoil \
                          + Cpool[istrlit] * k[istrlit] * frac_strlit_to_micsoil \
                          + Cpool[islowsoil] * k[islowsoil] * frac_slowsoil_to_micsoil \
                          + Cpool[ipasssoil] * k[ipasssoil] * frac_passsoil_to_micsoil \
                          - Cpool[imicsoil] * k[imicsoil]
        dC_dt[islowsoil] = Cpool[imicsoil] * k[imicsoil] * frac_micsoil_to_slowsoil \
                           + Cpool[istrlit] * k[istrlit] * frac_strlit_to_slowsoil \
                           - Cpool[islowsoil] * k[islowsoil]
        dC_dt[ipasssoil] = Cpool[imicsoil] * k[imicsoil] * frac_micsoil_to_passsoil \
                           + Cpool[islowsoil] * k[islowsoil] * frac_slowsoil_to_passsoil \
                           - Cpool[ipasssoil] * k[ipasssoil]

        for i in range(len(Cpool)):
            Cpool[i] += (dC_dt[i] * dt)


def matrix_diagnostic():
    global diag_I, diag_ResidenceTime, diag_Capacity, diag_Potential

    # X(t) = -(AK)**(-1)BI(t) + (AK)**(-1) dX/dt = Xc(t) - Xp(t)
    # calculate AK
    matrix_AK = np.matmul(matrix_A, matrix_K)

    # a.) I = Cinput
    diag_I = Cinput

    try:    # try to take inverse of matrix_AK (otherwise/except matrix is invertible)
        # calculate (AK)**-1
        matrix_AKinv = inverse(matrix_AK)

        # b.) ResidenceTime
        diag_ResidenceTime = -(np.matmul(matrix_AKinv, matrix_B))

        # c.) C storage capacity
        # Xc(t) = -(AK)**(-1)BI(t) * dt
        diag_Capacity = -(np.matmul(matrix_AKinv, matrix_B)) * Cinput

        # d.) C storage potential
        # Xp(t) = -(AK)**(-1) dX/dt * dt
        diag_Potential = -(np.matmul(matrix_AKinv, dC_dt))

    except:  # matrix_AK is invertible, and resulting values are not a number (np.nan or 'nan')
        diag_ResidenceTime = [np.nan for r in range(npools)]      # 'nan' displayed as #NUM! in output Excel worksheet
        diag_Capacity = [np.nan for c in range(npools)]
        diag_Potential = [np.nan for p in range(npools)]


nacc = 0
matrix_AK = [[0 for col in range(npools)] for r in range(npools)]
matrix_BI = [0 for bi in range(npools)]


def matrix_spinup():
    global nacc, Cpool, matrix_AK, matrix_BI
    matrix_AK = np.add(matrix_AK, (np.matmul(matrix_A, matrix_K)))
    matrix_BI = np.add(matrix_BI, (np.multiply(Cinput, matrix_B)))
    nacc += 1
    if is_endof_forcing():
        matrix_AK = np.divide(matrix_AK, nacc)
        matrix_BI = np.divide(matrix_BI, nacc)
        try:        # try to take inverse of matrix_AK (otherwise/except matrix is invertible)
            matrix_AKinv = inverse(matrix_AK)
            Cpool = -(np.matmul(matrix_AKinv, matrix_BI))
        except:     # matrix_AK is invertible, and Cpool values are not a number (np.nan or 'nan')
            Cpool = [np.nan for c in range(npools)]       # 'nan' displayed as #NUM! in output Excel worksheet
        nacc = 0
        matrix_AK = [[0 for col in range(npools)] for r in range(npools)]
        matrix_BI = [0 for bi in range(npools)]


row = 1     # write output starting in second row (since first row is headers)
nave = 0

Cinput_out = 0
diag_ResidenceTime_out = 0
diag_I_out = 0
diag_Capacity_out = 0
diag_Potential_out = 0
Cpool_out = [0 for o in range(npools)]


def write_output():
    global row, nave, Cinput_out, diag_ResidenceTime_out, diag_I_out, diag_Capacity_out, diag_Potential_out, Cpool_out

    if not is_diagnostic:
        diag_I_out = -9999                 # -9999 is some very small number
        diag_ResidenceTime_out = -9999
        diag_Capacity_out = -9999
        diag_Potential_out = -9999

    if is_startof_year():
        nave = 0
        Cinput_out = 0
        Cpool_out = [0 for u in range(npools)]
        if is_diagnostic:
            diag_I_out = 0
            diag_ResidenceTime_out = 0
            diag_Capacity_out = 0
            diag_Potential_out = 0

    if is_write_yearly:  # write output yearly
        if is_endof_year():
            is_write = True
        else:
            is_write = False
    else:                # write output daily
        is_write = True

    nave += 1
    Cinput_out += Cinput

    if np.nan in Cpool:                                # if any element of Cpool is 'nan'
        Cpool_out = [np.nan for c in range(npools)]    # then all elements of Cpool_out are 'nan,' or not a number
    else:
        Cpool_out = np.add(Cpool_out, Cpool)           # matrix addition

    if is_diagnostic:
        diag_I_out += diag_I
        if np.nan in diag_ResidenceTime:
            diag_ResidenceTime_out = np.nan
        else:
            diag_ResidenceTime_out += np.sum(diag_ResidenceTime)
        if np.nan in diag_Capacity:
            diag_Capacity_out = np.nan
        else:
            diag_Capacity_out += np.sum(diag_Capacity)
        if np.nan in diag_Potential:
            diag_Potential_out = np.nan
        else:
            diag_Potential_out += np.sum(diag_Potential)

    if is_write:
        Cinput_out /= nave
        if np.nan not in Cpool_out:
            for i in range(len(Cpool_out)):
                Cpool_out[i] /= nave
        if is_diagnostic:
            diag_I_out = (diag_I_out / nave) * 3600 * 24 * 365
            if diag_ResidenceTime_out != np.nan:
                diag_ResidenceTime_out = (((diag_ResidenceTime_out / nave) / 3600) / 24) / 365
            # else diag_ResidenceTime_out stays np.nan
            if diag_Capacity_out != np.nan:
                diag_Capacity_out /= nave
            if diag_Potential_out != np.nan:
                diag_Potential_out /= nave

        # writing to Excel spreadsheet
        worksheet.write(row, 0, year, cell_format)
        worksheet.write(row, 1, doy, cell_format)
        worksheet.write(row, 2, Cinput_out, cell_format)
        worksheet.write(row, 3, Cpool_out[ifoliage], cell_format)
        worksheet.write(row, 4, Cpool_out[iwoody], cell_format)
        worksheet.write(row, 5, Cpool_out[imetlit], cell_format)
        worksheet.write(row, 6, Cpool_out[istrlit], cell_format)
        worksheet.write(row, 7, Cpool_out[imicsoil], cell_format)
        worksheet.write(row, 8, Cpool_out[islowsoil], cell_format)
        worksheet.write(row, 9, Cpool_out[ipasssoil], cell_format)
        if diag_I_out == np.inf:                          # if diag_I_out is infinity (divided by zero), 9999 (a very
            worksheet.write(row, 10, 9999, cell_format)   # big number) will be displayed -- instead of "infinity" or
        else:                                             # error #DIV/0!
            worksheet.write(row, 10, diag_I_out, cell_format)
        worksheet.write(row, 11, diag_ResidenceTime_out, cell_format)
        worksheet.write(row, 12, diag_Capacity_out, cell_format)
        worksheet.write(row, 13, diag_Potential_out, cell_format)
        row += 1    # go down to next row


open_io_file()

for iyr in range(1, (nyr_sim * 365) + 1):   # plus one ( + 1) because of upper bound exclusivity
    C_input()
    update_ecosysC()
    if is_diagnostic:
        matrix_diagnostic()
    if is_SASU_spinup:
        matrix_spinup()
    write_output()
    tick_time()

workbook.close()    # close Excel workbook
